﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using TFDP.Common;
using System.ComponentModel.Composition;
using Microsoft.Data.Schema.SchemaModel;
using Microsoft.Data.Schema;
using TFDP.Common.Configuration;
using System.ComponentModel.Composition.Hosting;
using TFDP.Common.T4;

namespace TFDP.SimpleProcessor
{
    [Export(typeof(IProcessor))]
    public sealed class Processor : IProcessor
    {
        #region Class Members
        private ThreadsafeQueue<IModelElement> workItems;
        private IExtenderType properties;
        private ThreadsafeQueue<GeneratedArtifact> generatedItems;
        private bool isRunning;
        private bool isBusy;
        private string projectName;
        private UserInteractionServices interactionServices;
        private SettingsCollection settings; 
        #endregion

        #region Construction/Initialization
        public Processor()
        {
            workItems = new ThreadsafeQueue<IModelElement>();
            generatedItems = new ThreadsafeQueue<GeneratedArtifact>();
            settings = new SettingsCollection()
            {
                new Setting<TemplateMappings>() { Key = "Mappings", Value = new TemplateMappings() }
            };
        }
        #endregion

        #region Properties
        [ImportMany(typeof(IGenerator))]
        public IEnumerable<IGenerator> Generators { get; private set; } 
        #endregion

        #region IProcessor Members
        public int Count
        {
            get { return workItems.Count; }
        }

        public IEnumerable<Type> SupportedTypes
        {
            get
            {
                return new Type[] { typeof(IModelElement) };
            }
        }

        public IExtenderType PropertyExtenderType
        {
            get
            {
                if (properties == null)
                    properties = new SimpleProcessorProperties();

                return properties;
            }
        }

        public ThreadsafeQueue<Microsoft.Data.Schema.SchemaModel.IModelElement> Queue
        {
            get { return this.workItems; }
        }

        public bool IsRunning
        {
            get { return isRunning; }
        }

        public bool IsStopped
        {
            get { return !isRunning; }
        }

        public bool IsBusy
        {
            get { return isBusy; }
        }

        public string ProjectName
        {
            get { return projectName; }
            private set
            {
                if (projectName != value)
                {
                    projectName = value;

                    RaisePropertyChanged("ProjectName");
                }
            }
        }

        public Common.Configuration.SettingsCollection Settings
        {
            get { return settings; }
        }

        public void Initialize(Microsoft.VisualStudio.Data.Schema.Package.Project.IDatabaseProjectNode project, IServiceProvider serviceProvider)
        {
            if (project == null)
                throw new ArgumentNullException("project");
            if (serviceProvider == null)
                throw new ArgumentException("serviceProvider");

            this.interactionServices = project.UserInteractionServices;
            
            ProjectName = project.Caption;

            project.ProjectSaved += new EventHandler<Microsoft.VisualStudio.Data.Schema.Package.Project.ProjectEventArgs>(project_ProjectSaved);

            LoadSettingsFromProject(project);

            // MEF Composition
            // Create a catalog for the assembly that we're part of
            var catalog = new AssemblyCatalog(this.GetType().Assembly);
            var container = new CompositionContainer(catalog);

            container.ComposeParts(this);

            // Initialize the generators
            foreach (IGenerator generator in Generators)
            {
                generator.Initialize(this, project);
                generator.TemplateRecompile += new EventHandler<TemplateRecompileEventArgs>(generator_TemplateRecompile);
            }

            // We can start now
            this.isRunning = false;
        }

        private void LoadSettingsFromProject(Microsoft.VisualStudio.Data.Schema.Package.Project.IDatabaseProjectNode project)
        {
            // Load settings
            SettingsCollection persistedSettings = project.GetProcessorSettings();
            if (persistedSettings.Count > 0)
                settings = persistedSettings;
        }

        private void project_ProjectSaved(object sender, Microsoft.VisualStudio.Data.Schema.Package.Project.ProjectEventArgs e)
        {
            LoadSettingsFromProject(e.DatabaseProjectNode);
        }

        public void Enqueue(Microsoft.Data.Schema.SchemaModel.IModelElement element)
        {
            if (element == null)
                throw new ArgumentNullException("element");

            // Copy the list to make sure that the deferred execution
            // of the LINQ statement does not throw an exception
            // that the list has changed
            List<IModelElement> currentItems = this.workItems.ToList();

            // Only enqueue an element if it is not already included
            // in the workItems queue
            var matches = from e in currentItems
                          where e.IsNameEqual(element)
                          select e;

            if (matches.Count() == 0)
            {
                this.workItems.Enqueue(element);

                RaiseQueueChanged();
            }
        }

        public GeneratedArtifact PopGenerationItem()
        {
            if (generatedItems.Count > 0)
                return generatedItems.Dequeue();
            else
                return null;
        }

        public void Start()
        {
            if (!this.isRunning)
            {
                this.isRunning = true;
                this.isBusy = false;
                RaisePropertyChanged("IsRunning");
                RaisePropertyChanged("IsStopped");
            }
        }

        public void Stop()
        {
            if (this.isRunning)
            {
                this.isRunning = false;
                this.isBusy = false;
                RaisePropertyChanged("IsRunning");
                RaisePropertyChanged("IsStopped");
            }
        }

        public void DoWork()
        {
            if (!isRunning || isBusy)
                return;

            while (workItems.Count > 0)
                ProcessWorkItems(0);
        }

        public string GetTypeName(Microsoft.Data.Schema.SchemaModel.IModelElement element)
        {
            if (element == null)
                throw new ArgumentNullException("element");

            return this.interactionServices.GetElementTypeDescription(element.ElementClass);
        }

        public event EventHandler<EventArgs> BeginProcessing;

        public event EventHandler<EventArgs> ProcessingComplete;

        public event EventHandler<EventArgs> GenerationCompleted;

        public event EventHandler<EventArgs> QueueChanged;

        public event EventHandler<TemplateRecompileEventArgs> TemplateRecompile;

        private void RaiseGenerationCompleted()
        {
            if (GenerationCompleted != null)
                GenerationCompleted(this, EventArgs.Empty);
        }

        private void RaiseBeginProcessing()
        {
            if (BeginProcessing != null)
                BeginProcessing(this, EventArgs.Empty);
        }

        private void RaiseProcessingComplete()
        {
            if (ProcessingComplete != null)
                ProcessingComplete(this, EventArgs.Empty);
        }

        private void RaiseQueueChanged()
        {
            if (QueueChanged != null)
                QueueChanged(this, EventArgs.Empty);
        }

        private void RaiseTemplateRecompile(TemplateRecompileEventArgs e)
        {
            if (TemplateRecompile != null)
            {
                TemplateRecompile(this, e);
            }
        }
        #endregion

        #region INotifyPropertyChanged Members

        public event System.ComponentModel.PropertyChangedEventHandler PropertyChanged;

        private void RaisePropertyChanged(string propertyName)
        {
            if (PropertyChanged != null)
                PropertyChanged(this, new System.ComponentModel.PropertyChangedEventArgs(propertyName));
        }
        #endregion

        #region Processing
        /// <summary>
        /// Process the items in the queue.
        /// To process all the items in the queu specify count as 0
        /// </summary>
        /// <param name="count">The number of items to process. 0 to process all the items in the queue.</param>
        private void ProcessWorkItems(int count)
        {
            if (!isRunning)
                return;

            // Set busy
            isBusy = true;

            if (count == 0)
                count = workItems.Count;

            int current = 0;
            int generationCount = 0;

            // Keep processing until count items have been
            // processed or the queue is exhausted.
            while (current++ < count && workItems.Count > 0)
            {
                IModelElement item = workItems.Dequeue();

                RaiseQueueChanged();

                ProcessWorkItem(item);

                generationCount++;

                if (generationCount > 10)
                {
                    generationCount = 0;

                    RaiseGenerationCompleted();
                }
            }

            // Set no busy
            isBusy = false;

            if (generationCount > 0)
                RaiseGenerationCompleted();
        }

        /// <summary>
        /// Process a single workItem. Run all the generators that support
        /// the workItem type.
        /// </summary>
        /// <param name="workItem">An IModelElement instance to process</param>
        private void ProcessWorkItem(IModelElement workItem)
        {
            if (workItem == null)
                throw new ArgumentNullException("workItem");

            if (!isRunning)
                throw new InvalidOperationException("We're not running and you should not be here!");

            foreach (IGenerator generator in GeneratorsForType(workItem.GetType()))
            {
                try
                {
                    IEnumerable<GeneratedArtifact> items = generator.Generate(workItem);

                    foreach (GeneratedArtifact item in items)
                        generatedItems.Enqueue(item);
                }
                catch (Exception ex)
                {
                    Logger.LogError("Generation failed for " + generator.GetType().Name + ": " + ex.Message);
                }
            }
        }

        /// <summary>
        /// Find all the generators that support the given type.
        /// </summary>
        /// <param name="targetType">The type to support</param>
        /// <returns>A List<GeneratorBase> with the supporting generators</returns>
        private List<IGenerator> GeneratorsForType(Type targetType)
        {
            if (targetType == null)
                throw new ArgumentNullException("targetType");

            var matches = from g in this.Generators
                          where targetType.GetInterface(g.SupportedType.Name) != null
                          select g;

            return matches.ToList<IGenerator>();
        }
        #endregion

        private void generator_TemplateRecompile(object sender, TemplateRecompileEventArgs e)
        {
            RaiseTemplateRecompile(e);
        }
    }
}
