﻿using System;
using System.Diagnostics;
using System.Linq;
using Microsoft.VisualStudio;
using Microsoft.VisualStudio.Data.Schema.Package.Project;
using Microsoft.VisualStudio.Shell.Interop;
using Microsoft.VisualStudio.Data.Schema.Package.Internal;
using System.Collections.Generic;
using VSLangProj80;
using Microsoft.VisualStudio.TeamArchitect.ModelingProject;
using System.IO;
using Microsoft.VisualStudio.Shell;
using EnvDTE;

namespace TFDP.VisualStudioInterop
{
    public class VsInteropUtilities
    {
        public static Guid VS2010SqlDatabaseProject = new Guid("{c8d11400-126e-41cd-887f-60bd40844f9e}");

        /// <summary>
        /// Obtain the project filename and project type Guid from the given IVsHierarchy
        /// </summary>
        /// <param name="hierarchy">The IVsHierarchy to examine</param>
        public static void DumpProjectNameAndType(IVsHierarchy hierarchy)
        {
            IVsAggregatableProject aggProject = hierarchy as IVsAggregatableProject;

            string projectType = String.Empty;
            aggProject.GetAggregateProjectTypeGuids(out projectType);

            string projectName = GetFileNameFromHierarchy(hierarchy);

            Trace.WriteLine("Project filename: " + projectName + " project type: " + projectType);
        }

        /// <summary>
        /// Obtain the filename of the given IVsHierarchy
        /// </summary>
        /// <param name="hierarchy">The IVsHierarchy to examine</param>
        /// <returns>The filename or String.Empty if the filename could not be obtained</returns>
        public static string GetFileNameFromHierarchy(IVsHierarchy hierarchy)
        {
            string fileName = String.Empty;

            uint itemId = GetItemId(hierarchy);

            hierarchy.GetCanonicalName(itemId, out fileName);

            return fileName;
        }

        /// <summary>
        /// Obtain the Guid that identifies the project type of the given IVsHierarchy
        /// </summary>
        /// <param name="hierarchy">The IVsHierarchy to examine</param>
        /// <returns>The project type Guid or Guid.Empty if the project type could not be obtained</returns>
        public static Guid GetProjectType(IVsHierarchy hierarchy)
        {
            string projectType = String.Empty;

            IVsAggregatableProject aggProject = hierarchy as IVsAggregatableProject;

            if (aggProject != null)
                aggProject.GetAggregateProjectTypeGuids(out projectType);

            return projectType != String.Empty ? new Guid(projectType) : Guid.Empty;
        }

        /// <summary>
        /// Determine if the project specified by hierarchy is a SQL 2008 Database project
        /// </summary>
        /// <param name="hierarchy">The IVsHierarchy to examine</param>
        /// <returns>true if hierarchy is a database project, otherwise false</returns>
        public static bool IsSqlDatabaseProject(IVsHierarchy hierarchy)
        {
            /// TODO: Include other SQL Database project Guids that can be supported
            return GetProjectType(hierarchy) == VS2010SqlDatabaseProject;
        }

        /// <summary>
        /// Obtain the item id of the IVsHierarchy
        /// </summary>
        /// <param name="hierarchy"></param>
        /// <returns></returns>
        public static uint GetItemId(IVsHierarchy hierarchy)
        {
            object extObject;
            uint itemId = 0;
            IVsHierarchy tempHierarchy;

            hierarchy.GetProperty(VSConstants.VSITEMID_ROOT, (int)__VSHPROPID.VSHPROPID_BrowseObject, out extObject);

            IVsBrowseObject browseObject = extObject as IVsBrowseObject;
            if (browseObject != null)
                browseObject.GetProjectItem(out tempHierarchy, out itemId);

            return itemId;
        }

        /// <summary>
        /// Obtain the IVsHierarchy specified by the document cookie from the RunningDocumentTable
        /// </summary>
        /// <param name="table">The IVsRunningDocumentTable to search</param>
        /// <param name="cookie">The document cookie</param>
        /// <returns>A IVsHierarchy that points to the document</returns>
        public static IVsHierarchy GetHierarchyFromDocCookie(IVsRunningDocumentTable table, uint cookie)
        {
            uint pgrfRDTFlags;
            uint pdwReadLocks;
            uint pdwEditLocks;
            string pbstrMkDocument;
            IVsHierarchy hierarchy;
            uint pitemid;
            IntPtr ppunkDocData;

            table.GetDocumentInfo(
                cookie,
                out pgrfRDTFlags,
                out pdwReadLocks,
                out pdwEditLocks,
                out pbstrMkDocument,
                out hierarchy,
                out pitemid,
                out ppunkDocData
            );

            return hierarchy;
        }

        /// <summary>
        /// Obtain the database project from the given IVsHierarchy
        /// </summary>
        /// <param name="hierarchy">The IVsHierarchy that points to the project root</param>
        /// <returns>A IDatabaseProjectNode instance or null if hierarchy is not a Database Project</returns>
        public static IDatabaseProjectNode GetDatabaseProject(IVsHierarchy hierarchy)
        {
            return hierarchy as IDatabaseProjectNode;
        }

        /// <summary>
        /// Obtain the EnvDTE.Project of a given IVsHierarchy
        /// </summary>
        /// <param name="hierarchy">The IVsHierarchy to examine</param>
        /// <returns>A EnvDTE.Project instance that is the parent of the IVsHierarchy</returns>
        public static EnvDTE.Project GetEnvDTEProject(IVsHierarchy hierarchy)
        {
            object prjObject = null;

            hierarchy.GetProperty((uint)VSConstants.VSITEMID.Root, (int)__VSHPROPID.VSHPROPID_ExtObject, out prjObject);

            return (EnvDTE.Project)prjObject;
        }

        /// <summary>
        /// Obtain the IDatabaseNode from the specified IVsHierarchy
        /// </summary>
        /// <param name="hierarchy">The IVsHierarchy that points to a IDatabaseNode</param>
        /// <returns>A IDatabaseNode instance or null if the hierarchy does not contain a IDatabaseNode</returns>
        public static IDatabaseNode GetNode(IVsHierarchy hierarchy)
        {
            object prjObject = null;

            hierarchy.GetProperty(
                (uint)VSConstants.VSITEMID.Selection,
                (int)__VSHPROPID.VSHPROPID_ExtObject,
                out prjObject
            );

            if (prjObject == null)
                return null;
            else if (prjObject is OAFileItem)
            {
                if (((OAFileItem)prjObject).Object is IDatabaseNode)
                    return (((OAFileItem)prjObject).Object as IDatabaseNode);
            }
            else if (prjObject is OAFolderItem)
            {
                if (((OAFolderItem)prjObject).Object is IDatabaseFolderNode)
                    return (((OAFolderItem)prjObject).Object as IDatabaseFolderNode);
            }

            return null;
        }

        /// <summary>
        /// Get the string representation of the given tagVSQueryEditResult
        /// </summary>
        /// <param name="result">A tagVSQueryEditResult value</param>
        /// <returns>The string representation of result or String.Empty</returns>
        public static string QueryEditResultToString(tagVSQueryEditResult result)
        {
            string retval = String.Empty;

            switch (result)
            {
                case tagVSQueryEditResult.QER_EditNotOK:
                    retval = "Edit has been disallowed";
                    break;
                case tagVSQueryEditResult.QER_EditOK:
                    retval = "Edit is allowed";
                    break;
                default:
                    break;
            }

            return retval;
        }

        /// <summary>
        /// Get the string representation of the given tagVSQueryEditResultFlags
        /// </summary>
        /// <param name="flags">A tagVSQueryEditResultFlags value</param>
        /// <returns>The string representation of flags or String.Empty</returns>
        public static string QueryEditResultFlagsToString(tagVSQueryEditResultFlags flags)
        {
            string retval = String.Empty;

            switch (flags)
            {
                case tagVSQueryEditResultFlags.QER_MaybeCheckedout:
                    retval = "Files checked-out to edit";
                    break;
                case tagVSQueryEditResultFlags.QER_MaybeChanged:
                    retval = "Files changed on checkout";
                    break;
                case tagVSQueryEditResultFlags.QER_InMemoryEdit:
                    retval = "Safe to edit files in memory";
                    break;
                case tagVSQueryEditResultFlags.QER_InMemoryEditNotAllowed:
                    retval = "Edit denied because in-memory edit not allowed";
                    break;
                case tagVSQueryEditResultFlags.QER_NoisyCheckoutRequired:
                    retval = "Silent mode operation does not permit UI";
                    break;
                // QER_NoisyPromptRequired seems to have a value that is shared
                // by another tagVSQueryEditResultFlag
                //case tagVSQueryEditResultFlags.QER_NoisyPromptRequired:
                //    retval = "Silent mode operation does not permit UI";
                //    break;
                case tagVSQueryEditResultFlags.QER_CheckoutCanceledOrFailed:
                    retval = "Edit not allowed because checkout failed";
                    break;
                case tagVSQueryEditResultFlags.QER_EditNotPossible:
                    retval = "Edit will never be allowed because of current option settings or external conditions";
                    break;
                case tagVSQueryEditResultFlags.QER_ReadOnlyNotUnderScc:
                    retval = "Edit not allowed because file is read-only on disk";
                    break;
                case tagVSQueryEditResultFlags.QER_ReadOnlyUnderScc:
                    retval = "Edit not allowed because file is read-only and under source control (probably checked in)";
                    break;
                default:
                    break;
            }

            return retval;
        }

        public static IEnumerable<IVsProject> ProjectsInSolution(IVsSolution solution)
        {
            IEnumHierarchies enumerator = null;
            Guid guid = Guid.Empty;
            solution.GetProjectEnum((uint)__VSENUMPROJFLAGS.EPF_LOADEDINSOLUTION, ref guid, out enumerator);
            IVsHierarchy[] hierarchy = new IVsHierarchy[1] { null };
            uint fetched = 0;
            for (enumerator.Reset(); enumerator.Next(1, hierarchy, out fetched) == VSConstants.S_OK && fetched == 1; /*nothing*/)
            {
                yield return (IVsProject)hierarchy[0];
            }
        }

        public static bool IsVsLangProject(IVsHierarchy hierarchy)
        {
            return (GetEnvDTEProject(hierarchy).Object is VSProject2);
        }

        public static VSProject2 GetVsLangProject(IVsProject projectHierarchy)
        {
            return GetEnvDTEProject((IVsHierarchy)projectHierarchy).Object as VSProject2;
        }

        public static bool IsModelingProject(IVsHierarchy hierarchy)
        {
            ModelingProjectNode project = GetModelingProject(hierarchy);
            
            return project != null;
        }

        public static ModelingProjectNode GetModelingProject(IVsHierarchy hierarchy)
        {
            return GetEnvDTEProject(hierarchy).Object as ModelingProjectNode;
        }

        public static IEnumerable<string> GetProjectsInSolution()
        {
            IVsSolution solutionService = Microsoft.VisualStudio.Shell.Package.GetGlobalService(typeof(IVsSolution)) as IVsSolution;

            if (solutionService != null)
            {
                var projects = VsInteropUtilities.ProjectsInSolution(solutionService);
                return projects
                    .Select(p => VsInteropUtilities.GetEnvDTEProject((IVsHierarchy)p))
                    .Select(p => p.Name)
                    .ToList();
            }

            return new List<string>();
        }

        public static IEnumerable<string> GetFilesInDirectory(
            string projectName, 
            string relativePath, 
            Func<string, bool> filter)
        {
            if (String.IsNullOrEmpty(projectName))
            {
                throw new ArgumentNullException("projectName");
            }

            IVsSolution solutionService = Microsoft.VisualStudio.Shell.Package.GetGlobalService(typeof(IVsSolution)) as IVsSolution;
            EnvDTE.Project project = null;

            if (solutionService == null)
            {
                return new List<string>();
            }

            var projects = VsInteropUtilities.ProjectsInSolution(solutionService);
            project = projects
                .Select(p => VsInteropUtilities.GetEnvDTEProject((IVsHierarchy)p))
                .Where(p => p.Name == projectName)
                .FirstOrDefault();

            if (project == null)
            {
                return new List<string>();
            }

            EnvDTE.ProjectItem folder = VsInteropUtilities.GetFolderProjectItem(project, relativePath);

            return folder
                .ProjectItems
                .OfType<EnvDTE.ProjectItem>()
                .Where(pi => filter(pi.Name))
                .Select(pi => pi.Name)
                .ToList();
        }

        public static string GetActiveProject()
        {
            DTE dte = Package.GetGlobalService(typeof(SDTE)) as DTE;
            
            Array activeSolutionProjects = dte.ActiveSolutionProjects as Array;

            EnvDTE.Project activeProject = null;

            if (activeSolutionProjects != null && activeSolutionProjects.Length > 0)
            {
                activeProject = activeSolutionProjects.GetValue(0) as Project;
            }

            return activeProject.Name;
        }

        private static EnvDTE.ProjectItem GetFolderProjectItem(EnvDTE.Project project, string relativePath)
        {
            string[] pathParts = relativePath.Split(
                new char[] { Path.DirectorySeparatorChar },
                StringSplitOptions.RemoveEmptyEntries
            );

            // Check if the target folders exist, starting at the root of the project
            EnvDTE.ProjectItem targetFolder = project.ProjectItems.Item(pathParts[0]) as EnvDTE.ProjectItem;

            // Index of ProjectItems starts at 1 not 0
            for (int i = 1; i < pathParts.Length; i++)
            {
                targetFolder = targetFolder.ProjectItems.Item(pathParts[i]);

                if (targetFolder == null)
                {
                    return null;
                }
            }

            return targetFolder;
        }
    }
}
